async def serving_endpoint_from_options(worker):
    host = worker.options.serving_host
    port = worker.options.serving_port
    endpoint = "http://{host}:{port}".format(host=host,port=port)
    return {"serving_endpoint": endpoint}


import logging
from aiohttp import web
import contextvars

from . import models
from . import exceptions

context_worker = contextvars.ContextVar('worker')


async def index(request):
    worker = context_worker.get()
    return web.json_response({
            "service": "Inferout Management API",
            "cluster_name": worker.cluster.name,
            "cluster_version": worker.cluster.version,
            "redis_key_prefix": worker.cluster.redis_key_prefix,
            "current_worker_id": worker.id
        }
        )

async def workers(request):
    worker = context_worker.get()
    return web.json_response({
            "workers" : await worker.get_all_workers_data()
        }
        )

async def namespaces(request):
    worker = context_worker.get()
    return web.json_response({
        "namespaces": await models.ModelNamespace.get_all_as_list(worker.cluster)
    })

async def namespace_get(request):
    worker = context_worker.get()

    id = request.match_info.get('id')
    ns = models.ModelNamespace(cluster=worker.cluster, id=id)
    try:
        await ns.read()
    except exceptions.NotFoundException:
        raise web.HTTPNotFound
    return web.json_response({
        "id": ns.id,
        "settings": ns.settings
    })

async def namespace_put(request):
    worker = context_worker.get()

    id = request.match_info.get('id')
    request_data = await request.json()
    ns = models.ModelNamespace(cluster=worker.cluster, id=id)
    ns.settings = request_data.get('settings')
    logging.info(request_data)
    await ns.save()
    return web.json_response({
        "id": ns.id,
        "settings": ns.settings
    })

async def model_get(request):
    worker = context_worker.get()

    id = request.match_info.get('id')
    namespace_id = request.match_info.get('namespace_id')

    ns = models.ModelNamespace(cluster=worker.cluster, id=namespace_id)
    try:
        await ns.read()
    except exceptions.NotFoundException:
        raise web.HTTPNotFound
    model = models.Model(namespace=ns, id=id)
    try:
        await model.read()
    except exceptions.NotFoundException:
        raise web.HTTPNotFound
    return web.json_response({
        "id": model.id,
        "parameters": model.parameters,
        "latest_version_id": model.latest_version_id
    })

async def model_put(request):
    worker = context_worker.get()

    id = request.match_info.get('id')
    namespace_id = request.match_info.get('namespace_id')
    request_data = await request.json()
    ns = models.ModelNamespace(cluster=worker.cluster, id=namespace_id)
    try:
        await ns.read()
    except exceptions.NotFoundException:
        raise web.HTTPNotFound

    model = models.Model(namespace=ns, id=id)
    try:
        await model.read()
    except exceptions.NotFoundException:#updating
        pass
    model.parameters = request_data.get('parameters', {})
    try:
        await model.save()
    except exceptions.NotFoundException:
        raise web.HTTPNotFound
    return web.json_response({
        "id": model.id,
        "parameters": model.parameters,
        "latest_version_id": model.latest_version_id
    })

async def models_index(request):
    worker = context_worker.get()
    namespace_id = request.match_info.get('namespace_id')

    ns = models.ModelNamespace(cluster=worker.cluster, id=namespace_id)
    try:
        await ns.read()
    except exceptions.NotFoundException:
        raise web.HTTPNotFound

    return web.json_response({
        "models": await models.Model.get_all_as_list(worker.cluster, ns)
    })

async def model_versions_index(request):
    worker = context_worker.get()
    namespace_id = request.match_info.get('namespace_id')
    model_id = request.match_info.get('model_id')

    ns = models.ModelNamespace(cluster=worker.cluster, id=namespace_id)
    try:
        await ns.read()
    except exceptions.NotFoundException:
        raise web.HTTPNotFound
    
    model = models.Model(namespace=ns, id=model_id)
    try:
        await model.read()
    except exceptions.NotFoundException:
        raise web.HTTPNotFound

    return web.json_response({
        "model_versions": await models.ModelVersion.get_all_as_list(worker.cluster, model)
    })

async def model_instances_index(request):
    worker = context_worker.get()
    namespace_id = request.match_info.get('namespace_id')
    model_id = request.match_info.get('model_id')

    ns = models.ModelNamespace(cluster=worker.cluster, id=namespace_id)
    try:
        await ns.read()
    except exceptions.NotFoundException:
        raise web.HTTPNotFound
    
    model = models.Model(namespace=ns, id=model_id)
    try:
        await model.read()
    except exceptions.NotFoundException:
        raise web.HTTPNotFound

    return web.json_response({
        "model_instances": await models.ModelInstance.get_all_as_list(worker.cluster, model)
    })

app = web.Application()
app.add_routes([
    web.get('/', index),
    web.get('/workers', workers),
    web.get('/namespaces', namespaces),
    web.get('/namespaces/{id}', namespace_get),
    web.put('/namespaces/{id}', namespace_put),
    web.get('/namespaces/{namespace_id}/models', models_index),
    web.get('/namespaces/{namespace_id}/models/{id}', model_get),
    web.put('/namespaces/{namespace_id}/models/{id}', model_put),
    web.get('/namespaces/{namespace_id}/models/{model_id}/versions', model_versions_index),
    web.get('/namespaces/{namespace_id}/models/{model_id}/instances', model_instances_index)
])


import asyncio
from copy import deepcopy
import json
from os import name

from inferout.models import (
    Model,
    ModelInstance,
    ModelNamespace,
    ModelVersion)
import async_timeout
import logging
from functools import partial

from .cluster import Cluster
from . import utils
import signal
from . import management_api
from . import serving_api
from aiohttp import web
import re
from .storage_engines.base import StorageEngine
from .serving_engines.base import ServingEngine
from . import exceptions
from . import worker_annotators

WORKER_ANNOTATORS = [worker_annotators.serving_endpoint_from_options]

WORKER_KEY = '{{@worker-{}}}'
WORKER_KEY_REGEX = re.compile('{@worker-(.*)}$')
WORKER_REPORT_SLEEP_DURATION = 5
WORKER_REPORT_EXPIRE_MULTIPLIER = 2
WORKER_REPORT_EXPIRE_ADDITION = 10

class Worker(object):
    def __init__(self, cluster: Cluster, options) -> None:
        self.cluster = cluster
        self.options = options
        self.id = utils.get_uuid_as_string()
        self.worker_key = WORKER_KEY.format(self.id)
        self.state = None
        self.shutdown_requested = False
        self.storage_engines = {}
        self.serving_engines = {}

        self.serving_endpoint = None

        self.local_model_instances = {}
    
    async def report_forever(self):
        while True:
            await self.report_once()
            if self.shutdown_requested:
                break
            await asyncio.sleep(WORKER_REPORT_SLEEP_DURATION)

    async def report_once(self, send_events=False):
            from .scheduler import SCHEDULER_KEY
            logging.debug("Reporting %s", self.worker_key)
            data = {
                "state": self.state,
                "serving_endpoint": self.serving_endpoint,
                "available_storage_engines": json.dumps(list(self.storage_engines.keys())),
                "available_serving_engines": json.dumps(list(self.serving_engines.keys())),
                "model_instances_count": len(self.local_model_instances)
                }
            logging.debug("Reporting %s %s", self.worker_key, data)
            
            redis_key = self.cluster.get_redis_key(self.worker_key)
            async with self.cluster.redis.pipeline(transaction=True) as pipe:
                pipe.hset(
                    redis_key,
                    mapping=data
                    )
                pipe.expire(redis_key,
                WORKER_REPORT_SLEEP_DURATION*WORKER_REPORT_EXPIRE_MULTIPLIER
                )
                await pipe.execute()
            if send_events:
                event_data = {
                    "worker_id": self.id,
                    "state": self.state,
                    "model_instances_count": len(self.local_model_instances)
                    }
                await self.cluster.redis.publish(
                    self.cluster.get_redis_channel_key(
                        SCHEDULER_KEY
                    ),
                    message=json.dumps({
                            "event_type": "WORKER_UPDATE",
                            "event_data": event_data
                        })
                )
    
    async def get_worker_data_from_redis(self, redis_key):
        worker_data = await self.cluster.redis.hgetall(redis_key)
        worker_data["id"] = WORKER_KEY_REGEX.search(redis_key).group(1)
        worker_data["available_storage_engines"] = json.loads(worker_data["available_storage_engines"])
        worker_data["available_serving_engines"] = json.loads(worker_data["available_serving_engines"])
        worker_data["model_instances_count"] = int(worker_data["model_instances_count"])
        return worker_data

    async def get_all_workers_data(self):
        data = []
        async for each in self.cluster.redis.scan_iter(
            self.cluster.get_redis_key(WORKER_KEY.format('*'))
            ):
            data.append(await self.get_worker_data_from_redis(each))
        return data
    
    async def get_remote_worker_data(self, worker_id):
        return await self.get_worker_data_from_redis(
            self.cluster.get_redis_key(WORKER_KEY.format(worker_id)
        ))

    def load_serving_engines(self, engines: list):
        for each in engines:
            engine_module = __import__(each,fromlist=[each])
            if not issubclass(engine_module.ServingEngine, ServingEngine):
                raise ValueError("Invalid ServingEngine ",each)
            engine_instance = engine_module.ServingEngine()
            engine_instance.validate_engine_options()
            self.serving_engines[engine_module.__name__] = engine_instance
        logging.info("available serving engines: %s",",".join(self.serving_engines))

    def load_storage_engines(self, engines: list):
        for each in engines:
            engine_module = __import__(each,fromlist=[each])
            if not issubclass(engine_module.StorageEngine, StorageEngine):
                raise ValueError("Invalid StorageEngine ",each)
            engine_instance = engine_module.StorageEngine()
            engine_instance.validate_engine_options()
            self.storage_engines[engine_module.__name__] = engine_instance
        logging.info("available storages engines: %s",",".join(self.storage_engines))
    
    async def shutdown(self, sig, loop):
        logging.info("Shutting down gracefully, reson=%s",sig)
        self.state = "shutting_down"
        await self.report_once(send_events=True)

        self.shutdown_requested = True
        self.scheduler.shutdown_requested = True

        logging.info("watting for worker_pubsub_task")
        await self.worker_pubsub_task
        logging.info("watting for report_forever")
        await self.report_forever_task
        logging.info("waitting for scheduler")
        await self.scheduler_task
        logging.info("waitting for scheduler_pubsub_task")
        await self.scheduler_pubsub_task
        logging.info("Shutting down gracefully Completed")
    

    async def _get_model_instance_from_event_data(self, event_data):
        if(event_data["worker_id"]!=self.id):
            logging.error("worker_id missmatch, expected %s found %s", self.id, event_data["worker_id"])
            return
        namespace = ModelNamespace(cluster=self.cluster, id=event_data["namespace_id"])
        try:
            await namespace.read()
        except exceptions.NotFoundException:
            return
        model = Model(namespace=namespace, id=event_data["model_id"])
        try:
            await model.read()
        except exceptions.NotFoundException:
            return
        model_version = ModelVersion(model=model, id=event_data["model_version_id"])
        try:
            await model_version.read()
        except exceptions.NotFoundException:
            return
        model_instance = ModelInstance(model_version, id=event_data["model_instance_id"])
        try:
            await model_instance.read()
        except exceptions.NotFoundException:
            return
        return model_instance

    async def handle_model_instance_scheduled(self, event_data):
        model_instance = await self._get_model_instance_from_event_data(event_data)
        await self.activate_model_instance(model_instance)
    
    async def handle_terminate_model_instance(self, event_data):
        model_instance = await self._get_model_instance_from_event_data(event_data)
        local_model_instance = self.local_model_instances[model_instance.redis_key]
        await self.deactivate_model_instance(local_model_instance)
    


    async def channel_reader(self):
        while not self.shutdown_requested:
            try:
                async with async_timeout.timeout(1):
                    message = await self.worker_pubsub.get_message(ignore_subscribe_messages=True)
                    if message is not None:
                        logging.debug(f"(Worker Reader) Message Received: {message}")
                        data = json.loads(message["data"])
                        if data["event_type"] == "MODEL_INSTANCE_SCHEDULED":
                            await self.handle_model_instance_scheduled(data["event_data"])
                        elif data["event_type"] == "TERMINATE_MODEL_INSTANCE":
                            await self.handle_terminate_model_instance(data["event_data"])
                        else:
                            logging.info("Unknown event type %s, skipping", data["event_type"])

                    await asyncio.sleep(0.01)
            except asyncio.TimeoutError:
                pass
    
    async def activate_model_instance(self, model_instance: ModelInstance):
        ns = model_instance.model.namespace
        storage_engine = self.storage_engines[ns.settings["storage_engine"]]
        serving_engine = self.serving_engines[ns.settings["serving_engine"]]
        model_parameters = model_instance.model_version.parameters

        loop = asyncio.get_event_loop()

        model_instance.state = "initializing"
        await model_instance.save()
        await self.register_local_model_instance(model_instance)

        try:
            await loop.run_in_executor(None, partial(storage_engine.validate_model_parameters,
                model_parameters=model_parameters))
        except Exception as e:
            model_instance.state = "fetch_validation_error"
            model_instance.error_messages = [str(e)]
            await model_instance.save()
            await self.register_local_model_instance(model_instance)
            return
        storage_context = None
        try:
            storage_context = await loop.run_in_executor(None, partial(storage_engine.fetch_model,
            model_parameters=model_parameters))
        except Exception as e:
            model_instance.state = "fetch_error"
            model_instance.error_messages = [str(e)]
            await model_instance.save()
            await self.register_local_model_instance(model_instance)
            return
        model_instance.state = "loadding"
        model_instance.storage_context = storage_context
        await model_instance.save()
        await self.register_local_model_instance(model_instance)

        try:
            await loop.run_in_executor(None, partial(serving_engine.validate_model_parameters,
                model_parameters=model_parameters))
        except Exception as e:
            model_instance.state = "load_model_validation_error"
            model_instance.error_messages = [str(e)]
            await model_instance.save()
            await self.register_local_model_instance(model_instance)
            return
        serving_context = None
        worker_serving_context = None
        try:
            serving_context, worker_serving_context  = await loop.run_in_executor(None, partial(serving_engine.load_model,
            model_parameters=model_parameters,
            storage_context=storage_context)
            )
        except Exception as e:
            model_instance.state = "load_model_error"
            model_instance.error_messages = [str(e)]
            await model_instance.save()
            await self.register_local_model_instance(model_instance)
            return
        model_instance.state = "serving"
        model_instance.storage_context = storage_context
        model_instance.serving_context = serving_context
        model_instance.worker_serving_context = worker_serving_context
        await model_instance.save()
        await self.register_local_model_instance(model_instance)
    
    async def deactivate_model_instance(self, model_instance: ModelInstance):
        ns = model_instance.model.namespace
        model_parameters = model_instance.model_version.parameters

        storage_engine = self.storage_engines[ns.settings["storage_engine"]]
        serving_engine = self.serving_engines[ns.settings["serving_engine"]]

        loop = asyncio.get_event_loop()

        logging.debug("terminating model instance id=%s model_id=%s version=%s namespace=%s",
            model_instance.id, model_instance.model.id, model_instance.model_version_id, model_instance.model.namespace.id)
        model_instance.state = "terminating"
        await model_instance.save()
        await self.report_once(send_events=True)
        try:
            await loop.run_in_executor(None, partial(serving_engine.unload_model,
            model_parameters=model_parameters,
            storage_context=model_instance.storage_context,
            serving_context=model_instance.serving_context,
            worker_serving_context=model_instance.worker_serving_context
            ))
        except Exception as e:
            logging.exception(e)
            logging.error("error unloading model")
        try:
            await loop.run_in_executor(None, partial(storage_engine.clean_model,
            model_parameters=model_parameters,
            storage_context=model_instance.storage_context
            ))
        except Exception as e:
            logging.exception(e)
            logging.error("error cleaning storage for model")
        
        await self.deregister_local_model_instance(model_instance)

    
    async def register_local_model_instance(self, model_instance:ModelInstance):
        self.local_model_instances[model_instance.redis_key] = model_instance
        await self.report_once(send_events=True)
    
    async def deregister_local_model_instance(self, model_instance:ModelInstance):
        try:
            del self.local_model_instances[model_instance.redis_key]
        except KeyError:
            pass
        await self.report_once(send_events=True)
    
    async def apply_worker_annotators(self):
        annotations = {}
        for each in WORKER_ANNOTATORS:
            annotations.update(await each(self))
        self.serving_endpoint = annotations.get("serving_endpoint")

    async def run_forever(self):
        from .scheduler import Scheduler
        
        await self.apply_worker_annotators()

        management_api_runner = web.AppRunner(management_api.app)
        management_api.context_worker.set(self)
        await management_api_runner.setup()
        management_api_site = web.TCPSite(management_api_runner,
            self.options.management_host, self.options.management_port)
        await management_api_site.start()
        logging.info("Management API started, host=%s port=%d",self.options.management_host, self.options.management_port)

        serving_api_runner = web.AppRunner(serving_api.app)
        serving_api.context_worker.set(self)
        await serving_api_runner.setup()
        serving_api_site = web.TCPSite(serving_api_runner,
            self.options.serving_host, self.options.serving_port)
        await serving_api_site.start()
        logging.info("Serving API started, host=%s port=%d",self.options.serving_host, self.options.serving_port)

        loop = asyncio.get_event_loop()

        self.state = "initializing"
        await self.report_once(send_events=True)
        self.report_forever_task = loop.create_task(self.report_forever())
        await loop.run_in_executor(None, self.load_storage_engines, self.options.storage_engines)
        await loop.run_in_executor(None, self.load_serving_engines, self.options.serving_engines)
        
        self.state = "serving"
        await self.report_once(send_events=True)

        self.worker_pubsub = self.cluster.redis.pubsub()
        await self.worker_pubsub.subscribe(
            self.cluster.get_redis_channel_key(self.worker_key)
            )

        self.worker_pubsub_task = loop.create_task(self.channel_reader())
        self.scheduler = Scheduler(cluster=self.cluster, worker=self)
        self.scheduler_task = loop.create_task(self.scheduler.schedule_forever())
        self.scheduler_pubsub_task = loop.create_task(await self.scheduler.setup_pubsub())

        signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
        for s in signals:
            loop.add_signal_handler(
            s, lambda s=s: asyncio.create_task(self.shutdown(s, loop)))

        await asyncio.gather(
            self.report_forever_task,
            self.scheduler_task,
            self.worker_pubsub_task,
            self.scheduler_pubsub_task
        )
    
    async def do_infer(self, model_instance:ModelInstance, data: dict):
        local_instance = self.local_model_instances[model_instance.redis_key]
        ns = local_instance.model.namespace
        serving_engine = self.serving_engines[ns.settings["serving_engine"]]
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, partial(serving_engine.infer,
            model_parameters=local_instance.model_version.parameters,
            storage_context=local_instance.storage_context,
            serving_context=local_instance.serving_context,
            worker_serving_context=local_instance.worker_serving_context,
            data=deepcopy(data))
            )


from .worker import Worker
from .cluster import Cluster
import configargparse
import aioredis
import asyncio
import logging

async def main(options):
    redis = await aioredis.from_url(
        options.redis_url,
        encoding="utf-8",
        decode_responses=True)
    command = options.command
    cluster = Cluster(redis=redis,
        redis_key_prefix=options.redis_key_prefix,
        name=options.cluster_name)

    if command == "bootstrap_cluster":
        await cluster.bootstrap()
    elif command == "worker":
        await cluster.sync()
        logging.info("Joinning Cluster: name=%s version=%s", cluster.name, cluster.version)
        worker = Worker(cluster=cluster, options=options)
        await worker.run_forever()

def setup_logging(options):
    logging.basicConfig(
        level=getattr(logging,options.log_level),
        #format='%(asctime)s | %(name)s | [%(pathname)s %(funcName)s %(lineno)d] | %(levelname)s | %(message)s'
        )

def execute_from_command_line(argv=None):
    p = configargparse.ArgParser()
    p.add('command', choices=["bootstrap_cluster", "worker"], help='command')
    p.add('--cluster-name', required=True, help='name of the cluster')
    p.add('--redis-url', required=True, help='redis url')
    p.add('--redis-key-prefix', default="inferout", help='key prefix used for redis keys')
    p.add('--log-level', default="INFO", choices=["DEBUG","INFO","WARNING"], help='Logging Level')
    p.add('--management-host', default="0.0.0.0", help='listen host for management API')
    p.add('--management-port', default="9500", help='listen port for management API', type=int)
    p.add('--serving-host', default="0.0.0.0", help='listen host for serving API')
    p.add('--serving-port', default="9510", help='listen port for serving API', type=int)

    p.add('--storage-engines', nargs='+', default=['inferout.storage_engines.local_files'])
    p.add('--serving-engines', nargs='+', default=['inferout.serving_engines.echo'])
    options, _unknown = p.parse_known_args()
    setup_logging(options)
    asyncio.run(main(options))    


from . import base
import configargparse
import os


def dir_path(string):
    if os.path.isdir(string):
        return string
    else:
        raise NotADirectoryError(string)

class StorageEngine(base.StorageEngine):
    def __init__(self) -> None:
        pass
    
    def validate_engine_options(self):
        p = configargparse.ArgParser()

        p.add('--local-files-store-dir', default='/tmp/infer_models', type=dir_path, help='Directory location for local_files store')
        options, _unknown = p.parse_known_args()
        self.base_dir = options.local_files_store_dir
    
    def prepare(self):
        pass
    
    def validate_model_parameters(self, model_parameters):
        if not model_parameters.get("path"):
            raise ValueError("Path is required")
        final_path = os.path.join(self.base_dir, model_parameters.get("path"))
        if not os.path.exists(final_path):
            raise ValueError("Does Not exist:", final_path)

    
    def fetch_model(self, model_parameters) -> dict:
        final_path = os.path.join(self.base_dir, model_parameters["path"])
        return {"local_path": final_path}
        
    def clean_model(self, model_parameters:dict, storage_context:dict):
        pass


from typing import Tuple


class ServingEngine(object):
    def __init__(self) -> None:
        raise NotImplementedError()
    
    def validate_engine_options(self):
        raise NotImplementedError()
    
    def prepare(self):
        raise NotImplementedError()
    
    def validate_model_parameters(self, model_parameters:dict):
        raise NotImplementedError()
    
    def load_model(self, model_parameters:dict, storage_context:dict) -> Tuple[dict, dict]:#load_context
        raise NotImplementedError()
    
    def unload_model(self, model_parameters:dict, storage_context:dict, serving_context: dict, worker_serving_context:dict):
        raise NotImplementedError()
    
    def infer(self, model_parameters:dict, storage_context:dict, serving_context: dict, worker_serving_context:dict, data: dict) -> dict:
        raise NotImplementedError()


from inferout.models import ModelInstance
from typing import Tuple
from inferout import worker
import re
from . import base
import configargparse
import os


class EchoInstance(object):
    def __init__(self, lines) -> None:
        self.tokens = "\n".join(lines).split()

class ServingEngine(base.ServingEngine):
    def __init__(self) -> None:
        pass
    
    def validate_engine_options(self):
        p = configargparse.ArgParser()

        p.add('--echo-serving-file-name', type=str, help='example value')
        options, _unknown = p.parse_known_args()
        self.file_name = options.echo_serving_file_name
    
    def prepare(self):
        pass
    
    def validate_model_parameters(self, model_parameters):
        allow_echo = model_parameters.get("allow_echo")
        file_name = model_parameters.get("file_name") or self.file_name
        if not allow_echo:
            raise ValueError("allow_echo is required")
        if not file_name:
            raise ValueError("file_name is required")
    
    def load_model(self, model_parameters:dict, storage_context:dict) -> Tuple[dict, dict]:
        file_name = model_parameters.get("file_name") or self.file_name
        file_path = os.path.join(storage_context["local_path"], file_name)
        text_lines = open(file_path, 'r').readlines()
        echo_instance = EchoInstance(lines=text_lines)
        serving_context = {"echo_model_loaded": True}
        worker_serving_context = {"text_lines": text_lines, "echo_instance": echo_instance}
        return (serving_context, worker_serving_context)
        
    def unload_model(self, model_parameters:dict, storage_context:dict, serving_context: dict, worker_serving_context:dict):
        del worker_serving_context["echo_instance"].tokens
    
    def infer(self, model_parameters:dict, storage_context:dict, serving_context: dict, worker_serving_context:dict, data: dict) -> dict:
        data["tokens"] = worker_serving_context["echo_instance"].tokens
        return data


import json
import copy
import logging
import re
from .cluster import Cluster
from inferout import cluster, exceptions, utils

NAMESPACE_REGEX = re.compile("^[a-z][a-z0-9\-_]{2,9}$") #min len 3, max len 10
NAMESPACE_KEY = "{{@namespace-{}}}"
NAMESPACE_KEY_REGEX = re.compile('{@namespace-(.*)}$')

MODEL_REGEX = re.compile("^[a-z][a-z0-9\-_]{0,19}$") #min len 1, max len 20
MODEL_KEY = "{{@model-{}}}"
MODEL_KEY_REGEX = re.compile('{@model-(.*)}$')

MODEL_VERSION_KEY = "@model_version-{}"
MODEL_VERSION_KEY_REGEX = re.compile('@model_version-(.*)$')

MODEL_INSTANCE_KEY = "@model_instance-{}"
MODEL_INSTANCE_KEY_REGEX = re.compile('@model_instance-(.*)$')

MODEL_INSTANCE_TERMINATING_DURATION = 60

NAMESPACE_DEFAULT_SETTINGS = {
    "storage_engine": "inferout.storage_engines.local_files",
    "serving_engine": "inferout.serving_engines.echo",
    "instances_per_model": {
        "min": 1,
        "max": 4,
        "target": 2
    },
    "max_version_history": 10
}

class ModelNamespace(object):
    def __init__(self, cluster: Cluster, id:str) -> None:
        if NAMESPACE_REGEX.match(id) is None:
            raise ValueError("invalid Namespace name %s", id)
        self.cluster = cluster
        self.id = id
        self.settings = None

        self.redis_key = self.cluster.get_redis_key(NAMESPACE_KEY.format(self.id))

    async def read_from_redis(self):
        data = await self.cluster.redis.hgetall(
            self.redis_key
        )
        if not data:
            raise exceptions.NotFoundException()

        self.settings = json.loads(data["settings"])
    
    async def read(self):
        await self.read_from_redis()

    async def save_to_redis(self):
        data = {"settings": json.dumps(self.settings)}
        await self.cluster.redis.hset(
            self.redis_key,
            mapping=data
        )
    
    async def save(self):
        settings = copy.deepcopy(NAMESPACE_DEFAULT_SETTINGS)
        self.settings = utils.deep_update(settings, self.settings or {})

        await self.save_to_redis()
    
    @classmethod
    async def get_all_as_list(cls, cluster):
        data = []
        async for each in cluster.redis.scan_iter(
            cluster.get_redis_key(NAMESPACE_KEY.format('*'))
            ):
            each_data = await cluster.redis.hgetall(each)
            each_data["id"] = NAMESPACE_KEY_REGEX.search(each).group(1)
            each_data["settings"] = json.loads(each_data["settings"])
            data.append(each_data)
        return data

class Model(object):
    def __init__(self, namespace: ModelNamespace, id: str) -> None:
        if MODEL_REGEX.match(id) is None:
            raise ValueError("invalid Model name %s", id)
        self.id = id
        self.namespace = namespace
        self.cluster = namespace.cluster
        
        self.latest_version = None
        self.latest_version_id = None

        self.parameters = None

        self.redis_key = self.cluster.get_redis_key(
            self.namespace.id,
            MODEL_KEY.format(self.id))
    
    @classmethod
    async def get_all_as_list(cls, cluster:Cluster, namespace:ModelNamespace):
        data = []
        async for each in cluster.redis.scan_iter(
            cluster.get_redis_key(namespace.id, MODEL_KEY.format('*'))
            ):
            each_data = await cluster.redis.hgetall(each)
            each_data["id"] = MODEL_KEY_REGEX.search(each).group(1)
            each_data["parameters"] = json.loads(each_data["parameters"])
            each_data["latest_version_id"] = int(each_data["latest_version_id"])
            data.append(each_data)
        return data

    async def read_from_redis(self):
        data = await self.cluster.redis.hgetall(
            self.redis_key
        )
        if not data:
            raise exceptions.NotFoundException()

        self.latest_version_id = int(data["latest_version_id"])
        self.parameters = json.loads(data["parameters"])

    async def read(self):
        await self.read_from_redis()
        self.latest_version = ModelVersion(model=self, id=self.latest_version_id)
        await self.latest_version.read()

    async def save_to_redis(self):
        data = {
            "latest_version_id": self.latest_version_id,
            "parameters": json.dumps(self.parameters)
        }
        await self.cluster.redis.hset(
            self.redis_key,
            mapping=data
        )

    async def save(self):
        self.parameters = self.parameters or {}
        if self.latest_version_id is None:
            self.latest_version_id = 1
        else:
            self.latest_version_id += 1
        latest_version = ModelVersion(model=self, id=self.latest_version_id)
        latest_version.parameters = self.parameters
        await latest_version.save()
        self.latest_version = self.parameters
        await self.save_to_redis()

class ModelVersion(object):
    def __init__(self, model:Model, id:str) -> None:
        self.model = model
        self.cluster = model.cluster
        self.id = id
        self.parameters = None

        self.redis_key = self.cluster.get_redis_key(
            self.model.namespace.id,
            utils.covert_to_redis_slot(self.model.id),
            MODEL_VERSION_KEY.format(self.id))
        
    @classmethod
    async def get_all_as_list(cls, cluster:Cluster, model: Model):
        data = []
        async for each in cluster.redis.scan_iter(
            cluster.get_redis_key(
                model.namespace.id,
                utils.covert_to_redis_slot(model.id),
                MODEL_VERSION_KEY.format("*")
                )
            ):
            each_data = await cluster.redis.hgetall(each)
            each_data["id"] = int(MODEL_VERSION_KEY_REGEX.search(each).group(1))
            each_data["parameters"] = json.loads(each_data["parameters"])
            data.append(each_data)
        return sorted(data, key=lambda x: x["id"], reverse=True)

    async def read_from_redis(self):
        data = await self.cluster.redis.hgetall(
            self.redis_key
        )
        if not data:
            raise exceptions.NotFoundException()

        self.parameters = json.loads(data["parameters"])

    async def read(self):
        await self.read_from_redis()

    async def save_to_redis(self):
        data = {
            "parameters": json.dumps(self.parameters)
        }
        await self.cluster.redis.hset(
            self.redis_key,
            mapping=data
        )

    async def save(self):
        self.parameters = self.parameters or {}
        await self.save_to_redis()

class ModelInstance(object):
    def __init__(self, model_version:ModelVersion, id) -> None:
        self.id = id
        self.model_version = model_version
        self.storage_context = None
        self.serving_context = None # avaiable in redis
        self.worker_serving_context = None #only avaiable in assigned worker
        self.worker_id = None
        self.state = None
        self.error_messages = None

        self.model = model_version.model

        self.model_version_id = model_version.id

        self.cluster = model_version.cluster
    
        self.redis_key = self.cluster.get_redis_key(
            self.model.namespace.id,
            utils.covert_to_redis_slot(self.model.id),
            str(self.model_version.id),
            MODEL_INSTANCE_KEY.format(self.id))

    async def read_from_redis(self):
        data = await self.cluster.redis.hgetall(
            self.redis_key
        )
        if not data:
            raise exceptions.NotFoundException()
        self.worker_id = data["worker_id"]
        self.state = data["state"]
        self.storage_context = json.loads(data["storage_context"])
        self.serving_context = json.loads(data["serving_context"])
        self.error_messages = json.loads(data["error_messages"])

    async def read(self):
        await self.read_from_redis()


    async def save_to_redis(self):
        data = {
            "worker_id": self.worker_id,
            "state": self.state,
            "storage_context": json.dumps(self.storage_context),
            "serving_context": json.dumps(self.serving_context),
            "error_messages": json.dumps(self.error_messages)
        }
        await self.cluster.redis.hset(
            self.redis_key,
            mapping=data
        )
        if self.state == "terminating":
            ttl = await self.cluster.redis.ttl(self.redis_key)
            if ttl == -1:
                await self.cluster.redis.expire(
                    self.redis_key,
                    MODEL_INSTANCE_TERMINATING_DURATION
                )

    async def save(self):
        self.storage_context = self.storage_context or {}
        self.serving_context = self.serving_context or {}
        await self.save_to_redis()
    
    @classmethod
    async def get_all_as_list(cls, cluster:Cluster, model: Model, version:ModelVersion=None):
        data = []
        if version is None:
            version_id_key = "*" #any version
        else:
            version_id_key = version.id
        async for each in cluster.redis.scan_iter(
            cluster.get_redis_key(
                model.namespace.id,
                utils.covert_to_redis_slot(model.id),
                version_id_key,
                MODEL_INSTANCE_KEY.format("*")
                )
            ):
            each_data = await cluster.redis.hgetall(each)
            each_data["id"] = MODEL_INSTANCE_KEY_REGEX.search(each).group(1)
            each_data["model_version_id"] = utils.split_redis_key(each)[-2]
            each_data["worker_id"] = each_data["worker_id"]
            each_data["state"] = each_data["state"]
            each_data["storage_context"] = json.loads(each_data["storage_context"])
            each_data["serving_context"] = json.loads(each_data["serving_context"])
            each_data["error_messages"] = json.loads(each_data["error_messages"])
            data.append(each_data)
        return sorted(data, key=lambda x: x["model_version_id"], reverse=True)
    


class StorageEngine(object):
    def __init__(self) -> None:
        raise NotImplementedError()
    
    def validate_engine_options(self):
        raise NotImplementedError()
    
    def prepare(self):
        raise NotImplementedError()
    
    def validate_model_parameters(self, model_parameters):
        raise NotImplementedError()
    
    def fetch_model(self, model_parameters) -> dict:#returns model fetch_context
        raise NotImplementedError()
    
    def clean_model(self, model_parameters:dict, storage_context:dict):
        raise NotImplementedError()


import json
import logging
from aiohttp import web
import contextvars
import random

import aiohttp

from . import models
from . import exceptions

context_worker = contextvars.ContextVar('worker')


async def index(request):
    worker = context_worker.get()

    return web.json_response({
            "service": "Inferout Serving API",
            "cluster_name": worker.cluster.name,
            "cluster_version": worker.cluster.version,
            "redis_key_prefix": worker.cluster.redis_key_prefix,
            "current_worker_id": worker.id
        }
        )

async def handle_infer_post(request):
    worker = context_worker.get()
    namespace_id = request.match_info.get('namespace_id')
    model_id = request.match_info.get('model_id')
    version_id = request.match_info.get('version_id')

    try:
        request_data = await request.json()
    except json.decoder.JSONDecodeError:
        raise web.HTTPBadRequest(text="Invalid JSON")
        
    input_data = request_data.get("input_data") or {}

    ns = models.ModelNamespace(cluster=worker.cluster, id=namespace_id)
    try:
        await ns.read()
    except exceptions.NotFoundException:
        raise web.HTTPNotFound

    model = models.Model(namespace=ns, id=model_id)
    try:
        await model.read()
    except exceptions.NotFoundException:#updating
        pass
    
    version = None
    if version_id in ("_latest", "latest", None):
        version = models.ModelVersion(model=model, id=str(model.latest_version_id))
    else:
        version = models.ModelVersion(model=model, id=version_id)
    try:
        await version.read()
    except exceptions.NotFoundException:#updating
        pass

    instances_data = await models.ModelInstance.get_all_as_list(
        cluster=worker.cluster,
        model=model,
        version=version)
    
    instances_data = list(filter(lambda x: x["state"]=="serving", instances_data))

    output_data = None

    try:
        local_instance_data = list(filter(lambda x: x["worker_id"]==worker.id, instances_data))[0]
    except IndexError:
        local_instance_data = None

    
    if local_instance_data:
        logging.info("Serving from local worker")
        instance = models.ModelInstance(model_version=version,id=local_instance_data["id"])
        output_data = await worker.do_infer(model_instance=instance, data=input_data)
        return web.json_response({
            "model_version": version.id,
            "worker_id": local_instance_data["worker_id"],
            "input_data": input_data,
            "output_data": output_data
            })
    else:
        random.shuffle(instances_data)
        for each in instances_data:
            remote_worker_data = await worker.get_remote_worker_data(worker_id=each["worker_id"])
            logging.info("Routing to remote worker id=%s endpoint=%s",
                remote_worker_data["id"], remote_worker_data["serving_endpoint"])
            async with aiohttp.ClientSession() as session:
                async with session.post(
                    remote_worker_data["serving_endpoint"]+request.path_qs,
                    json=request_data) as response:
                    response_data = await response.json()
                    return web.json_response(response_data)
        
        logging.error("No active Model Instances found")
        raise web.HTTPServiceUnavailable

app = web.Application()
app.add_routes([
    web.get('/', index),
    web.post('/{namespace_id}/{model_id}', handle_infer_post),
    web.post('/{namespace_id}/{model_id}/{version_id}', handle_infer_post),
    ])


from distutils.core import setup
from setuptools import find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
  name = 'inferout',
  packages = find_packages(),
  version = '0.0.0a0',
  license='GPLv3',
  description = 'Distributed Scale Out Framework for ML models serving/inference',
  long_description=long_description,
  long_description_content_type="text/markdown",
  url = 'https://github.com/svrdev27/py-inferout',
  download_url = 'https://github.com/svrdev27/py-inferout/archive/v0.0.0-alpha.0.tar.gz',
  keywords = ['distributed', 'scale', 'ML', 'models', 'serving', 'inference', 'framework'],
  install_requires=[
          'aioredis',
          'aiohttp',
          'ConfigArgParse'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7'
  ],
  entry_points = {
        'console_scripts': ['inferout=inferout:execute_from_command_line']
    }
)



import aioredis
from . import utils
from .exceptions import InvalidClusterError
import logging

CLUSTER_BOOTSTRAP_VERSION = "0.0"

REDIS_LOCK_TIMEOUT = 10
REDIS_LOCK_BLOCKING_TIMEOUT = 0

CLUSTER_INFO_KEY = '@cluster_info'


class Cluster(object):
    def __init__(self, redis: aioredis.client.Redis, redis_key_prefix: str, name: str) -> None:
        self.redis = redis
        self.redis_key_prefix = redis_key_prefix
        self.name = name
        self.version = None
    
    def get_async_lock(self, *args):
        return self.redis.lock(
            name=utils.get_redis_key(self.redis_key_prefix, 'lock', *args),
            timeout=REDIS_LOCK_TIMEOUT,
            blocking_timeout=REDIS_LOCK_BLOCKING_TIMEOUT,
            thread_local=False)
    
    def get_redis_key(self, *args):
        return utils.get_redis_key(self.redis_key_prefix, *args)
    
    def get_redis_channel_key(self, *args):
        return utils.get_redis_key(self.redis_key_prefix, 'channel', *args)
    
    async def sync(self):
        cluster_info = await self.redis.hgetall(self.get_redis_key(CLUSTER_INFO_KEY))
        if not cluster_info:
            raise InvalidClusterError("Does Not exist.")
        if cluster_info['name'] != self.name:
            raise InvalidClusterError("Name mismatch")
        self.version = cluster_info['version']


    async def bootstrap(self):
        async with self.get_async_lock(CLUSTER_INFO_KEY):
            existing_cluster_info = await self.redis.hgetall(self.get_redis_key(CLUSTER_INFO_KEY))
            if(existing_cluster_info):
                logging.error("Existing cluster found")
                logging.info("Cluster Info: %s", existing_cluster_info)
            else:
                cluster_info = {
                    "name": self.name,
                    "version": CLUSTER_BOOTSTRAP_VERSION
                    }
                await self.redis.hset(self.get_redis_key(CLUSTER_INFO_KEY),
                mapping=cluster_info)
                logging.info("Bootstrapping Cluster: %s", cluster_info)


import uuid
import base64
import collections.abc

def deep_update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d

def get_redis_key(*args) -> str:
    return "::".join(args)

def split_redis_key(key) -> list:
    return key.split("::")


def get_uuid_as_string() -> str:
    return base64.urlsafe_b64encode(uuid.uuid4().bytes).rstrip(b'=').decode('ascii')

def covert_to_redis_slot(key:str) -> str:
    return "{{{}}}".format(key)



import asyncio
from asyncio.tasks import sleep
from copy import deepcopy
import json
import time

import aioredis
import async_timeout
from . import exceptions

from configargparse import Namespace
from inferout.management_api import workers
import logging

from .cluster import Cluster
from . import utils
from .storage_engines.base import StorageEngine
from .serving_engines.base import ServingEngine
from .worker import (Worker, WORKER_KEY)
from .models import (
    MODEL_INSTANCE_KEY,
    ModelNamespace,
    Model,
    ModelVersion,
    ModelInstance,
    MODEL_KEY,
    MODEL_KEY_REGEX,
    MODEL_INSTANCE_KEY_REGEX)
from . import utils

SCHEDULER_INTERVAL = 5
SCHEDULER_SLEEP_WARNNING_DURATION = 3
SCHEDULER_LOCK_ERROR_SLEEP_DURATION = 0.5
SCHEDULER_KEY = "@scheduler"
MODEL_INSTANCE_SCHEDULED_SLEEP_DURATION = 0.1

class Scheduler(object):
    def __init__(self, cluster: Cluster, worker: Worker) -> None:
        self.cluster = cluster
        self.worker = worker
        self.shutdown_requested = False

        self.active_workers_map = {}

    async def select_worker(self, model_version: ModelVersion):
        storage_engine = model_version.model.namespace.settings.get("storage_engine")
        serving_engine = model_version.model.namespace.settings.get("serving_engine")
        workers = self.active_workers_map.values()
        if storage_engine:
            workers = filter(lambda x: storage_engine in x["available_storage_engines"], workers)
        if serving_engine:
            workers = filter(lambda x: serving_engine in x["available_serving_engines"], workers)
        
        if not workers:
            return None
        workers = sorted(workers, key=lambda x: int(x["model_instances_count"]))
        return workers[0]["id"]
    
    async def refresh_active_workers_data(self):
        active_workers_map = {}
        for each_worker in await self.worker.get_all_workers_data():
            if each_worker["state"]=="serving":
                active_workers_map[each_worker["id"]] = each_worker
        self.active_workers_map = active_workers_map

    async def schedule_once(self):
        cluster = self.cluster
        async for each in cluster.redis.scan_iter(
            cluster.get_redis_key(
                "*",#any namespace
                MODEL_KEY.format("*"),#any model
                )
            ):

            ns_id, model_key = utils.split_redis_key(each)[1:]
            model_id = MODEL_KEY_REGEX.search(model_key).group(1)
            ns = ModelNamespace(cluster=cluster, id=ns_id)
            try:
                await ns.read()
            except exceptions.NotFoundException:
                continue
            model = Model(namespace=ns,id=model_id)
            try:
                await model.read()
            except exceptions.NotFoundException:
                continue
            logging.debug("looking for pending model instances of namespace=%s model=%s version=%s", ns.id, model.id, model.latest_version_id)
            latest_model_instances = []
            outdated_model_instances = []
            async for each_instance_key in cluster.redis.scan_iter(
                cluster.get_redis_key(
                    ns.id,
                    utils.covert_to_redis_slot(model.id),
                    "*", #any version
                    MODEL_INSTANCE_KEY.format("*") #any instance
                    )
                ):
                logging.debug("each_instance_key %s", each_instance_key)
                version_id, instance_key = utils.split_redis_key(each_instance_key)[-2:]
                version_id = int(version_id)
                instance_id = MODEL_INSTANCE_KEY_REGEX.search(instance_key).group(1)

                model_version = ModelVersion(model=model, id=version_id)
                try:
                    await model_version.read()
                except exceptions.NotFoundException:
                    continue

                model_instance = ModelInstance(model_version=model_version,id=instance_id)
                try:
                    await model_instance.read()
                except exceptions.NotFoundException:
                    continue
                if model_instance.state != "terminating" and (model_instance.worker_id not in self.active_workers_map):
                    model_instance.state = "terminating"
                    await model_instance.save()
                    continue
                if model.latest_version_id == model_instance.model_version_id:
                    latest_model_instances.append(model_instance)
                else:
                    outdated_model_instances.append(model_instance)
            logging.debug("outdated_model_instances: %s", ",".join([x.id for x in outdated_model_instances]))
            logging.debug("latest_model_instances: %s", ",".join([x.id for x in latest_model_instances]))
            
            latest_serving_model_instances = list(filter(lambda x: x.state=="serving",latest_model_instances))
            if len(latest_serving_model_instances) >= int(ns.settings["instances_per_model"]["target"]):
                for outdated_model_instance in outdated_model_instances:
                    if outdated_model_instance.state == "terminating":
                        continue
                    event_data = {
                        "namespace_id": ns.id,
                        "model_id": outdated_model_instance.model.id,
                        "model_version_id": outdated_model_instance.model_version.id,
                        "model_instance_id": outdated_model_instance.id,
                        "worker_id": outdated_model_instance.worker_id
                        }
                    await self.cluster.redis.publish(
                        self.cluster.get_redis_channel_key(
                            WORKER_KEY.format(outdated_model_instance.worker_id)
                        ),
                        message=json.dumps({
                                "event_type": "TERMINATE_MODEL_INSTANCE",
                                "event_data": event_data
                            })
                    )
                    logging.info("terminating model instance %s", event_data)

            no_new_instances_required = int(ns.settings["instances_per_model"]["target"]) - len(latest_model_instances)
            logging.debug("New model instances required: %d.", no_new_instances_required)

            for i in range(no_new_instances_required):
                new_model_instance = ModelInstance(
                    model_version=model.latest_version,
                    id=utils.get_uuid_as_string())
                new_model_instance.worker_id = await self.select_worker(
                    model_version=model.latest_version)
                if new_model_instance.worker_id is None:
                    logging.error("No sutable workers available for namespace=%s model=%s version=%s",ns.id,model.id, model.latest_version_id)
                    break
                new_model_instance.state = "scheduled"
                await new_model_instance.save()
                event_data = {
                    "namespace_id": ns.id,
                    "model_id": new_model_instance.model.id,
                    "model_version_id": new_model_instance.model_version.id,
                    "model_instance_id": new_model_instance.id,
                    "worker_id": new_model_instance.worker_id
                    }
                await self.cluster.redis.publish(
                    self.cluster.get_redis_channel_key(
                        WORKER_KEY.format(new_model_instance.worker_id)
                    ),
                    message=json.dumps({
                            "event_type": "MODEL_INSTANCE_SCHEDULED",
                            "event_data": event_data
                        })
                )
                await asyncio.sleep(MODEL_INSTANCE_SCHEDULED_SLEEP_DURATION)
                logging.info("scheduled model instance %s", event_data)

    async def handle_worker_update(self, event_data):
        logging.debug("handle_worker_update %s", event_data)
        worker_id = event_data["worker_id"]
        worker_data = await self.worker.get_remote_worker_data(worker_id=worker_id)
        if worker_data["state"] == "serving":
            self.active_workers_map[worker_id] = worker_data
        else:
            if worker_id in self.active_workers_map:
                del self.active_workers_map[worker_id]

    async def channel_reader(self):
        while not self.shutdown_requested:
            try:
                async with async_timeout.timeout(1):
                    message = await self.pubsub.get_message(ignore_subscribe_messages=True)
                    if message is not None:
                        logging.debug(f"(Scheduler Reader) Message Received: {message}")
                        data = json.loads(message["data"])
                        if data["event_type"] == "WORKER_UPDATE":
                            await self.handle_worker_update(data["event_data"])
                        else:
                            logging.info("Unknown event type %s, skipping", data["event_type"])

                    await asyncio.sleep(0.01)
            except asyncio.TimeoutError:
                pass
    
    async def setup_pubsub(self):
        self.pubsub = self.cluster.redis.pubsub()
        await self.pubsub.subscribe(
            self.cluster.get_redis_channel_key(SCHEDULER_KEY)
            )
        return self.channel_reader()

    async def schedule_forever(self):
        while not self.shutdown_requested:
            try:
                async with self.cluster.get_async_lock(SCHEDULER_KEY):
                    start_time = time.time()
                    await self.refresh_active_workers_data()
                    await self.schedule_once()
                    end_time = time.time()
                    time_to_sleep = SCHEDULER_INTERVAL - (end_time - start_time)
                    if time_to_sleep <= 0:
                        raise Exception("scheduler taking unexpectedly longer")
                    if time_to_sleep <= SCHEDULER_SLEEP_WARNNING_DURATION:
                        logging.warning("scheduler taking longer, took %f, sleeping for %f",(end_time - start_time), time_to_sleep)
                    logging.debug("sleeping for %f",time_to_sleep)
                    await asyncio.sleep(time_to_sleep)
            except aioredis.exceptions.LockError:
                logging.debug("Unable to acquire lock for scheduler")
                await asyncio.sleep(SCHEDULER_LOCK_ERROR_SLEEP_DURATION)